﻿/*
 * Author: G. Monz (DK7IO), 2012-04-30
 * This file is distributed without any warranty.
 * */

using System;
using System.Text;
using System.Globalization;
using System.Collections.Generic;
using System.Xml;
using System.Windows.Forms;
using System.ComponentModel;
using System.IO;
using System.Net;
using RadioMobile;

namespace KML2NET
{
	public partial class MainForm : Form
	{
		OpenFileDialog _inputFileDialog = new OpenFileDialog();
		SaveFileDialog _outputFileDialog = new SaveFileDialog();

		public MainForm()
		{
			InitializeComponent();

			Text = "KML2NET (by DK7IO, 2012-04-30)";

			//File selectors
			{
				_inputFileDialog.Filter = "KML files (*.kml)|*.kml|All files (*.*)|*.*";
				_inputFileDialog.FilterIndex = 1;

				_outputFileDialog.Filter = "NET files (*.net)|*.net|All files (*.*)|*.*";
				_outputFileDialog.FilterIndex = 1;
			}

			//Elevation data from GeoNames.org
			{
				checkBoxElevationDataFromGeoNamesOrg.Checked = true;
			}

			//Info label
			{
				labelInfo.Text = null;
			}

			SetDefaults();
		}

		void SetDefaults()
		{
			//Parameters
			{
				textBoxNetName.Text = "Receiving sites";

				textBoxNetMinMHz.Text = (87.5).ToString();
				textBoxNetMaxMHz.Text = (108.0).ToString();

				textNetBoxSurfaceRefractivity.Text = (301).ToString();
				textNetBoxGroundConductivity.Text = (0.005).ToString();
				textNetBoxRelativeGroundPermittivity.Text = (15).ToString();

				radioButtonNetPolarizationHorizontal.Checked = true;

				radioButtonNetModeOfVariabilityBroadcast.Checked = true;
				textBoxNetModeOfVariabilityPercentOfTime.Text =
					textBoxNetModeOfVariabilityPercentOfLocations.Text =
					textBoxNetModeOfVariabilityPercentOfSituations.Text =
				(70).ToString();

				radioButtonNetClimateContinentalTemperate.Checked = true;
			}

			//System
			{
				Single watt = (Single)1000;

				textBoxSystemAntennaHeight.Text = (1).ToString();
				textBoxSystemTransmitPower.Text = (watt).ToString();
				textBoxSystemName.Text = watt.ToString() + " W, ND";
			}

			//Units
			{
				textBoxUnitsAntennaHeight.Text = (1).ToString();
				textBoxUnitsIcon.Text = (72).ToString();
			}
		}

		static object GetSelectedRadioButtonTag(Control parentControl)
		{
			object result = null;
			{
				Control.ControlCollection controls = parentControl.Controls;

				foreach (Control control in controls)
				{
					RadioButton radioButton = control as RadioButton;

					if (radioButton != null)
					{
						if (radioButton.Checked)
						{
							result = control.Tag;
							break;
						}
					}
				}
			}

			if (result != null)
			{
				return result;
			}
			else
			{
				throw new InvalidOperationException(parentControl.Name + ": No radio button selected.");
			}
		}

		NetPolarization GetNetPolarization()
		{
			object tag = GetSelectedRadioButtonTag(groupBoxNetPolarization);
			TypeConverter converter = TypeDescriptor.GetConverter(typeof(NetPolarization));
			NetPolarization result = (NetPolarization)converter.ConvertFrom(tag);

			return result;
		}

		NetClimate GetNetClimate()
		{
			object tag = GetSelectedRadioButtonTag(groupBoxNetClimate);
			TypeConverter converter = TypeDescriptor.GetConverter(typeof(NetClimate));
			NetClimate result = (NetClimate)converter.ConvertFrom(tag);

			return result;
		}

		NetModeOfVariability GetNetModeOfVariability()
		{
			object tag = GetSelectedRadioButtonTag(groupBoxNetModeOfVariability);
			TypeConverter converter = TypeDescriptor.GetConverter(typeof(NetModeOfVariability));
			NetModeOfVariability result = (NetModeOfVariability)converter.ConvertFrom(tag);

			return result;
		}

		IList<RMNet> GetNets()
		{
			IList<RMNet> result = new List<RMNet>();
			{
				RMNet item = new RMNet(textBoxNetName.Text);
				{
					item.MinimumFrequency = Utilities.ParseSingle(textBoxNetMinMHz);
					item.MaximumFrequency = Utilities.ParseSingle(textBoxNetMaxMHz);

					item.SurfaceRefractivity = Utilities.ParseSingle(textNetBoxSurfaceRefractivity);
					item.GroundConductivity = Utilities.ParseSingle(textNetBoxGroundConductivity);
					item.RelativeGroundPermittivity = Utilities.ParseSingle(textNetBoxRelativeGroundPermittivity);

					item.Polarization = GetNetPolarization();

					item.ModeOfVariability = GetNetModeOfVariability();
					item.PercentOfTime = Utilities.ParseSingle(textBoxNetModeOfVariabilityPercentOfTime);
					item.PercentOfLocations = Utilities.ParseSingle(textBoxNetModeOfVariabilityPercentOfLocations);
					item.PercentOfSituations = Utilities.ParseSingle(textBoxNetModeOfVariabilityPercentOfSituations);

					item.Climate = GetNetClimate();
				}
				result.Add(item);
			}

			return result;
		}

		IList<RMSystem> GetSystems()
		{
			IList<RMSystem> result = new List<RMSystem>();
			{
				RMSystem item = new RMSystem();
				{
					item.Name = textBoxSystemName.Text;
					item.TransmitPower = Utilities.ParseSingle(textBoxSystemTransmitPower);
					item.AntennaGain = (Single)(0); //non directional antenna
					item.AntennaHeight = Utilities.ParseSingle(textBoxSystemAntennaHeight);
				}
				result.Add(item);
			}

			return result;
		}

		IList<RMUnit> GetUnits(string kmlFile)
		{
			XmlDocument kmlDocument = new XmlDocument();
			kmlDocument.Load(kmlFile);

			XmlNamespaceManager xmlNamespaceManager = new XmlNamespaceManager(kmlDocument.NameTable);
			{
				xmlNamespaceManager.AddNamespace("kml", "http://www.opengis.net/kml/2.2");
				//xmlNamespaceManager.AddNamespace("gx", "http://www.google.com/kml/ext/2.2");
				//xmlNamespaceManager.AddNamespace("atom", "http://www.w3.org/2005/Atom");
			}

			IList<RMUnit> result = GetUnits(xmlNamespaceManager, kmlDocument);

			return result;
		}

		IList<RMUnit> GetUnits(XmlNamespaceManager xmlNamespaceManager, XmlNode kmlDocument)
		{
			IList<RMUnit> result;
			{
				XmlNodeList placemarks = kmlDocument.SelectNodes("//kml:Placemark", xmlNamespaceManager);
				result = GetUnits(xmlNamespaceManager, placemarks);
			}

			return result;
		}

		IList<RMUnit> GetUnits(XmlNamespaceManager xmlNamespaceManager, XmlNodeList placemarks)
		{
			IList<RMUnit> result = new List<RMUnit>();
			{
				foreach (XmlNode placemark in placemarks)
				{
					XmlNode point = placemark.SelectSingleNode("kml:Point", xmlNamespaceManager);

					if (point != null)
					{
						XmlNode coordinatesNode = point.SelectSingleNode("kml:coordinates", xmlNamespaceManager);

						if (coordinatesNode != null)
						{
							string coordinates = Utilities.TrimText(coordinatesNode.InnerText);

							if (!Utilities.IsEmpty(coordinates))
							{
								string name = string.Empty;
								{
									XmlNode nameNode = placemark.SelectSingleNode("kml:name", xmlNamespaceManager);

									if (nameNode != null)
									{
										string innerText = nameNode.InnerText;

										if (innerText != null)
										{
											name = nameNode.InnerText;
										}
									}
								}

								RMUnit unit;
								{
									string longitude;
									string latitude;
									string height;
									{
										char[] separators = new char[] { ',' };
										string[] splittedCoordinates = coordinates.Split(separators);

										longitude = splittedCoordinates[0];
										latitude = splittedCoordinates[1];
										height = splittedCoordinates[2];
									}
									unit = GetUnitByKmlCoordinates(name, longitude, latitude, height);
								}
								result.Add(unit);
							}
						}
					}
				}
			}

			return result;
		}

		RMUnit GetUnitByKmlCoordinates(string name, string kmlLongitude, string kmlLatitude, string kmlHeight)
		{
			RMUnit result = new RMUnit(name);
			{
				result.Longitude = Utilities.ParseSingle(kmlLongitude.Replace(".", ","));
				result.Latitude = Utilities.ParseSingle(kmlLatitude.Replace(".", ","));
				result.Elevation = Utilities.ParseSingle(kmlHeight.Replace(".", ","));
			}

			return result;
		}

		RMCoverage GetCoverage()
		{
			RMCoverage result = new RMCoverage();
			{
			}

			return result;
		}

		#region Input validation
		string GetInputError()
		{
			string separator = " -> ";

			string result = null;

			if (result == null)
			{
				result = GetInputErrorForString(textBoxKmlFile, groupBoxKmlFile.Text + separator + labelKmlFile.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForSingle(textBoxNetMinMHz, groupBoxNetParameters.Text + separator + labelNetMinMHz.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForSingle(textBoxNetMaxMHz, groupBoxNetParameters.Text + separator + labelNetMaxMHz.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForSingle(textNetBoxSurfaceRefractivity, groupBoxNetParameters.Text + separator + labelNetSurfaceRefractivity.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForSingle(textNetBoxGroundConductivity, groupBoxNetParameters.Text + separator + labelNetGroundConductivity.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForSingle(textNetBoxRelativeGroundPermittivity, groupBoxNetParameters.Text + separator + labelNetRelativeGroundPermittivity.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForSingle(textBoxNetModeOfVariabilityPercentOfTime, groupBoxNetParameters.Text + separator + labelNetModeOfVariabilityPercentOfTime.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForSingle(textBoxNetModeOfVariabilityPercentOfLocations, groupBoxNetParameters.Text + separator + labelNetModeOfVariabilityPercentOfLocations.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForSingle(textBoxNetModeOfVariabilityPercentOfSituations, groupBoxNetParameters.Text + separator + labelNetModeOfVariabilityPercentOfSituations.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForSingle(textBoxSystemAntennaHeight, groupBoxSystem.Text + separator + labelSystemAntennaHeight.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForSingle(textBoxSystemTransmitPower, groupBoxSystem.Text + separator + labelSystemTransmitPower.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForSingle(textBoxUnitsAntennaHeight, groupBoxUnits.Text + separator + labelUnitsAntennaHeight.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForByte(textBoxUnitsIcon, groupBoxUnits.Text + separator + labelUnitIcon.Text);
			}

			if (result == null)
			{
				result = GetInputErrorForString(textBoxNetFile, groupBoxNetFile.Text + separator + labelNetFile.Text);
			}

			return result;
		}

		string GetInputErrorForString(TextBoxBase textBoxBase, string descriptionControlText)
		{
			string result = null;
			{
				if (Utilities.IsEmpty(textBoxBase.Text, true))
				{
					result = "'" + descriptionControlText + "' is empty";
				}
			}

			return result;
		}

		string GetInputErrorForByte(TextBoxBase textBoxBase, string descriptionControlText)
		{
			string result = GetInputErrorForString(textBoxBase, descriptionControlText);

			if (result == null)
			{
				if (!Utilities.TryParseByte(textBoxBase))
				{
					result = "'" + descriptionControlText + "' is not valid";
				}
			}

			return result;
		}

		string GetInputErrorForSingle(TextBoxBase textBoxBase, string descriptionControlText)
		{
			string result = GetInputErrorForString(textBoxBase, descriptionControlText);

			if (result == null)
			{
				if (!Utilities.TryParseSingle(textBoxBase))
				{
					result = "'" + descriptionControlText + "' is not valid";
				}
			}

			return result;
		}
		#endregion

		string Save()
		{
			string kmlFileName = @textBoxKmlFile.Text;
			string netFileName = @textBoxNetFile.Text;

			RM_NetFile netFile = new RM_NetFile3000();
			{
				netFile.Nets = GetNets();

				netFile.Units = GetUnits(kmlFileName);

				if (checkBoxElevationDataFromGeoNamesOrg.Checked)
				{
					bool success;
					{
						labelInfo.Text = "Retrieving elevation data for " + netFile.Units.Count.ToString() + " units from GeoNames.org";
						Refresh();
						{
							success = FillUnitsElevationsFromGeoNamesOrg(netFile.Units);
						}
						labelInfo.Text = null;
						Refresh();
					}

					if (!success)
					{
						MessageBox.Show("GeoNames.org has returned some errors!", Text, MessageBoxButtons.OK, MessageBoxIcon.Warning);
					}
				}

				netFile.Systems = GetSystems();

				netFile.Coverage = GetCoverage();

				netFile.SetNetRoles((Byte)128);
				netFile.SetNetSystems((Byte)netFile.Systems.Count);
				netFile.SetNetAntennaHeights(Utilities.ParseSingle(textBoxUnitsAntennaHeight));
				netFile.SetUnitIcons(Utilities.ParseByte(textBoxUnitsIcon));
				netFile.SetAdditionalCableLossesPerMeter(0);
			}
			netFile.Save(netFileName);

			string result =
				netFileName + Environment.NewLine +
				"with " + netFile.Units.Count.ToString() + " units created."
			;

			return result;
		}

		bool FillUnitsElevationsFromGeoNamesOrg(IList<RMUnit> units)
		{
			const int maxPerRequest = 20;

			bool result = true;
			int count = units.Count;

			if (count > maxPerRequest)
			{
				IList<RMUnit> requestUnits = new List<RMUnit>();
				int start = 0;
				bool cancel = (count == 0);

				while (!cancel)
				{
					int end = start + maxPerRequest - 1;
					requestUnits.Clear();
					bool hasMore = true;

					for (int i = start; i <= end; i++)
					{
						if (i < count)
						{
							requestUnits.Add(units[i]);
						}
						else
						{
							hasMore = false;
							break;
						}
					}

					if (requestUnits.Count > 0)
					{
						result = FillUnitsElevationsFromGeoNamesOrg(requestUnits);

						if (
							result &&
							hasMore
						)
						{
							start += maxPerRequest;

							//Wait a little while between two requests
							{
								int milliseconds = 500 + new Random().Next(1000);
								System.Threading.Thread.Sleep(milliseconds);
							}
						}
						else
						{
							cancel = true;
						}
					}
					else
					{
						cancel = true;
					}
				}
			}
			else if (count > 0)
			{
				NumberFormatInfo numberFormatInfo = NumberFormatInfo.InvariantInfo;
				StringBuilder url = new StringBuilder("http://api.geonames.org/srtm3");

				{
					url.Append("?lats=");
					int i = 0;

					foreach (RMUnit unit in units)
					{
						url.Append(unit.Latitude.ToString(numberFormatInfo));

						if (++i < count)
						{
							url.Append(",");
						}
					}
				}

				{
					url.Append("&lngs=");
					int i = 0;

					foreach (RMUnit unit in units)
					{
						url.Append(unit.Longitude.ToString(numberFormatInfo));

						if (++i < count)
						{
							url.Append(",");
						}
					}
				}

				url.Append("&username=demo");

				HttpWebRequest request = (HttpWebRequest)WebRequest.Create(url.ToString());
				HttpWebResponse response = (HttpWebResponse)request.GetResponse();
				StreamReader reader = new StreamReader(response.GetResponseStream());

				int counter = 0;
				string meters;

				while ((meters = reader.ReadLine()) != null)
				{
					meters = meters.Trim();

					if (meters.Contains("<"))
					{
						//HTML content

						result = false;
						Utilities.StartProcess(url.ToString(), null, null);
						break;
					}
					else
					{
						if (meters != "-32768")
						{
							units[counter].Elevation = Utilities.ParseSingle(meters);
						}
					}

					counter++;
				}
			}

			return result;
		}

		#region GUI events
		void buttonKmlFile_Click(object sender, EventArgs e)
		{
			if (_inputFileDialog.ShowDialog() == DialogResult.OK)
			{
				textBoxKmlFile.Text = _inputFileDialog.FileName;

				if (Utilities.IsEmpty(textBoxNetFile.Text, true))
				{
					textBoxNetFile.Text = Path.ChangeExtension(textBoxKmlFile.Text, "net");
				}
			}
		}

		void buttonNetFile_Click(object sender, EventArgs e)
		{
			if (_outputFileDialog.ShowDialog() == DialogResult.OK)
			{
				textBoxNetFile.Text = _outputFileDialog.FileName;
			}
		}

		void buttonStart_Click(object sender, EventArgs e)
		{
			try
			{
				string inputError = GetInputError();

				if (inputError == null)
				{
					string message = Save();
					MessageBox.Show(message, Text, MessageBoxButtons.OK, MessageBoxIcon.Information);
				}
				else
				{
					MessageBox.Show(inputError, Text, MessageBoxButtons.OK, MessageBoxIcon.Stop);
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}
		#endregion
	}
}